"""
Examples for differential equations systems with dynamic control.

Here we provide several systems whose state is governed by differential
equations that allow us to plug in a controller. The controller receives
the state of the system as input and its output is added to one (or more) of
the differential equations.
"""



"""A cubic controller."""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __cubic_2d_1o(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a cubic polynomial for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s02: Final[float] = s0 * s0
    s12: Final[float] = s1 * s1
    out[0] = (s0 * params[0]) + (s1 * params[1]) \
        + (s02 * params[2]) + (s0 * s1 * params[3]) + (s12 * params[4]) \
        + (s02 * s0 * params[5]) + (s02 * s1 * params[6]) \
        + (s0 * s12 * params[7]) + (s12 * s1 * params[8])


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __cubic_3d_1o(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a cubic polynomial for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    s02: Final[float] = s0 * s0
    s12: Final[float] = s1 * s1
    s22: Final[float] = s2 * s2
    out[0] = (s0 * params[0]) + (s1 * params[1]) + (s2 * params[2]) \
        + (s02 * params[3]) + (s0 * s1 * params[4]) \
        + (s0 * s2 * params[5]) + (s12 * params[6]) \
        + (s1 * s2 * params[7]) + (s22 * params[8]) \
        + (s0 * s02 * params[9]) + (s02 * s1 * params[10]) \
        + (s02 * s2 * params[11]) + (s12 * s1 * params[12]) \
        + (s12 * s0 * params[13]) + (s12 * s2 * params[14]) \
        + (s22 * s2 * params[15]) + (s22 * s0 * params[16]) \
        + (s22 * s1 * params[17])


def cubic(system: System) -> Controller:
    """
    Create a cubic controller for the given equations object.

    :param system: the equations object
    :return: the cubic controller
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    name: Final[str] = "cubic"
    if system.state_dims == 2:
        return Controller(name, 2, 1, 9, __cubic_2d_1o)
    if system.state_dims == 3:
        return Controller(name, 3, 1, 19, __cubic_3d_1o)
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
An alternative objective function for the dynamic control problem.

Different from :mod:`~moptipyapps.dynamic_control.objective`, we compute
the mean of `log(z + 1)` where `z` be the figures of merit of the single
training cases. We then return `exp(mean[log(z + 1)]) - 1` as final
result. The goal is to reduce the impact of training cases that require
more control effort.

If we solve the dynamic control problem for diverse training cases, then we
may have some very easy cases, where the system just needs a small control
impulse to move into a stable and cheap state. Others may have very far out
and expensive starting states that require lots of control efforts to be
corrected. If we simply average over all states, then these expensive states
will dominate whatever good we are doing in the cheap states. Averaging over
the `log(z+1)` reduces such impact. We then compute `exp[...]-1` of the result
as cosmetics to get back into the original range of the figure of merits.
"""

from math import expm1
from typing import Callable, Final

import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error

from moptipyapps.dynamic_control.instance import Instance
from moptipyapps.dynamic_control.ode import j_from_ode, run_ode
from moptipyapps.shared import SCOPE_INSTANCE


class FigureOfMeritLE(Objective):
    """Compute a `exp(mean(log(z+1)))-1` over the figures of merit `z`."""

    def __init__(self, instance: Instance) -> None:
        """
        Create the figure-of-merit objective of the dynamic control problem.

        :param instance: the instance
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the dynamic control instance
        self.instance: Final[Instance] = instance

        #: the simulation steps
        self.__steps: Final[int] = instance.system.training_steps
        #: the training starting states
        self.__training: Final[np.ndarray] = \
            instance.system.training_starting_states
        #: the results
        self.__results: Final[np.ndarray] = np.empty(
            len(self.__training), float)
        #: the equations
        self.__equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] = \
            instance.system.equations
        #: the controller
        self.__controller: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] = \
            instance.controller.controller
        #: the controller dimension
        self.__controller_dim: Final[int] = instance.controller.control_dims

    def __str__(self) -> str:
        """
        Get the name of this objective function.

        :return: `figureOfMeritLE`
        """
        return "figureOfMeritLE"

    def evaluate(self, x: np.ndarray) -> float:
        """
        Evaluate the parameterization of a controller.

        :param x: the controller parameters
        :return: the figure of merit
        """
        steps: Final[int] = self.__steps
        training: Final[np.ndarray] = self.__training
        results: Final[np.ndarray] = self.__results
        equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] \
            = self.__equations
        controller: Final[
            Callable[[np.ndarray, float, np.ndarray, np.ndarray], None]] \
            = self.__controller
        controller_dim: Final[int] = self.__controller_dim

        for i, start in enumerate(training):
            results[i] = j_from_ode(run_ode(
                start, equations, controller, x, controller_dim, steps),
                len(start))
        return expm1(np.log1p(results).mean())

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as scope:
            self.instance.log_parameters_to(scope)

    def lower_bound(self) -> float:
        """
        Get the lower bound of the figure of merit, which is 0.

        :returns: 0.0
        """
        return 0.0



"""Classes for representing a problem to be solved."""

from typing import Any, Callable, Final, Iterable, TypeVar

import moptipy.utils.plot_utils as pu
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import array_to_str
from moptipy.utils.path import Path
from moptipy.utils.types import check_to_int_range, type_error

from moptipyapps.dynamic_control.ode import multi_run_ode
from moptipyapps.dynamic_control.results_log import ResultsLog
from moptipyapps.dynamic_control.results_plot import ResultsPlot

#: the color for training points
_TRAINING_COLOR: Final[str] = "orange"
#: the color for testing points
_TEST_COLOR: Final[str] = "cyan"

#: the type variable for ODE controller parameters
T = TypeVar("T")


def _dummy_ctrl(_: np.ndarray, __: float, ___: Any,
                out: np.ndarray) -> None:
    """Do nothing as internal dummy controller that does nothing."""
    out.fill(0.0)


class System(Component):
    """A class for governing a system via differential system."""

    def __init__(self, name: str, state_dims: int, control_dims: int,
                 test_starting_states: np.ndarray,
                 training_starting_states: np.ndarray,
                 test_steps: int = 5000,
                 training_steps: int = 1000,
                 plot_examples: Iterable[int] = (0, )) -> None:
        """
        Initialize the system.

        :param name: the name of the system.
        :param state_dims: the state dimensions
        :param control_dims: the control dimensions
        :param test_starting_states: the starting states to be used for
            testing, as a matrix with one point per row
        :param test_steps: the steps to be taken for the test simulation
        :param training_steps: the steps to be taken for the training
            simulation
        :param training_starting_states: the starting states to be used for
            training, as a matrix with one point per row
        :param plot_examples: the points that should be plotted
        """
        super().__init__()
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        #: the name of the model
        self.name: Final[str] = name
        #: the dimensions of the state variable
        self.state_dims: Final[int] = check_to_int_range(
            state_dims, "state_dims", 2, 3)
        #: the dimensions of the controller output
        self.control_dims: Final[int] = check_to_int_range(
            control_dims, "control_dims", 1, 100)

        if not isinstance(test_starting_states, np.ndarray):
            raise type_error(test_starting_states, "test_starting_states",
                             np.ndarray)
        if (len(test_starting_states) < 1) or (
                test_starting_states.shape[1] != state_dims):
            raise ValueError(
                f"invalid test starting states {test_starting_states!r}.")
        #: the test starting states
        self.test_starting_states: Final[np.ndarray] = test_starting_states

        if not isinstance(training_starting_states, np.ndarray):
            raise type_error(training_starting_states,
                             "training_starting_states", np.ndarray)
        if (len(training_starting_states) < 1) or (
                training_starting_states.shape[1] != state_dims):
            raise ValueError("invalid training starting "
                             f"states {training_starting_states!r}.")
        #: the test starting states
        self.training_starting_states: Final[np.ndarray] = \
            training_starting_states

        #: the test simulation steps
        self.test_steps: Final[int] = check_to_int_range(
            test_steps, "test_steps", 10, 1_000_000_000)
        #: the training simulation steps
        self.training_steps: Final[int] = check_to_int_range(
            training_steps, "training_steps", 10, 1_000_000_000)

        #: the plot examples
        self.plot_examples: Final[tuple[int, ...]] = tuple(sorted(set(
            plot_examples)))
        total: Final[int] = len(self.training_starting_states) \
            + len(self.test_starting_states) - 1
        for i in self.plot_examples:
            check_to_int_range(i, "plot_examples[i]", 0, total)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("stateDims", self.state_dims)
        logger.key_value("controlDims", self.control_dims)
        logger.key_value("testStartingStates",
                         array_to_str(self.test_starting_states.flatten()))
        logger.key_value(
            "trainingStartingStates",
            array_to_str(self.training_starting_states.flatten()))
        logger.key_value("testSteps", self.test_steps)
        logger.key_value("trainingSteps", self.training_steps)
        logger.key_value("examplePlots", array_to_str(np.array(
            self.plot_examples, int).flatten()))

    def __str__(self):
        """
        Get the name of this model.

        :return: the name of this model
        """
        return self.name

    def equations(self, state: np.ndarray, time: float,
                  control: np.ndarray, out: np.ndarray) -> None:
        """
        Compute the values of the differential equations to be simulated.

        :param state: the state of the system
        :param time: the time index
        :param control: the output of the controller
        :param out: the differential, i.e., the output of this function
        """

    def plot_points(self, dest_dir: str, skip_if_exists: bool = True) -> Path:
        """
        Plot the training and testing points of the equation.

        :param dest_dir: the destination directory
        :param skip_if_exists: `True` to skip existing files, `False` otherwise
        :return: the plotted file
        """
        name: Final[str] = f"{self.name}_points"
        dest_folder: Final[Path] = Path.path(dest_dir)
        dest_folder.ensure_dir_exists()
        dest_file: Final[Path] = dest_folder.resolve_inside(f"{name}.pdf")
        if dest_file.ensure_file_exists() and skip_if_exists:
            return dest_file

        figure = pu.create_figure()
        if self.state_dims == 3:
            axes = figure.add_subplot(projection="3d")
            axes.set_aspect("equal")
            axes.force_zorder = True
            axes.set_zlabel("z")

            axes.scatter(self.training_starting_states[:, 0],
                         self.training_starting_states[:, 1],
                         self.training_starting_states[:, 2],
                         color=_TRAINING_COLOR, marker="x",
                         zorder=1)
            axes.scatter(self.test_starting_states[:, 0],
                         self.test_starting_states[:, 1],
                         self.test_starting_states[:, 2],
                         color=_TEST_COLOR, marker="o", zorder=2)
            axes.scatter(0, 0, 0, color="black", marker="+", zorder=3)
            zlim = axes.get_zlim()
            axes.set_zlim(1.1 * zlim[0], 1.1 * zlim[1])
        elif self.state_dims == 2:
            axes = pu.get_axes(figure)
            axes.set_aspect("equal")
            axes.grid(True)
            axes.scatter(self.training_starting_states[:, 0],
                         self.training_starting_states[:, 1],
                         color=_TRAINING_COLOR, marker="x",
                         zorder=1)
            axes.scatter(self.test_starting_states[:, 0],
                         self.test_starting_states[:, 1],
                         color=_TEST_COLOR, marker="o", zorder=2)
        else:
            raise ValueError(f"invalid dimensions {self.state_dims} "
                             f"for {self}.")

        xlim = axes.get_xlim()
        axes.set_xlim(1.1 * xlim[0], 1.1 * xlim[1])
        ylim = axes.get_ylim()
        axes.set_ylim(1.1 * ylim[0], 1.1 * ylim[1])
        axes.set_title(f"{self.name} system\ntraining points ("
                       f"{_TRAINING_COLOR}), test points ({_TEST_COLOR})")

        axes.set_xlabel("x")
        axes.set_ylabel("y")
        res = pu.save_figure(figure, name, dest_folder, "pdf")[0]
        if res != dest_file:
            raise ValueError(
                f"Wrong destination file {res!r}!={dest_file!r}?")
        return dest_file

    def describe_system(
            self, title: str | None,
            controller: Callable[[np.ndarray, float, T, np.ndarray], None],
            parameters: T, base_name: str, dest_dir: str,
            skip_if_exists: bool = False) \
            -> tuple[Path, Path]:
        """
        Describe the performance of a given system of system.

        :param title: the title
        :param controller: the controller to simulate
        :param parameters: the controller parameters
        :param base_name: the base name of the file to produce
        :param dest_dir: the destination directory
        :param skip_if_exists: if the file already exists
        :returns: the paths of the generated files
        """
        dest: Final[Path] = Path.path(dest_dir)
        dest.ensure_dir_exists()

        the_title = f"{self.name} system"
        if title is not None:
            the_title = f"{the_title}: {title}"

        file_1 = dest.resolve_inside(f"{base_name}_plot.pdf")
        file_2 = dest.resolve_inside(f"{base_name}_results.csv")
        if file_1.ensure_file_exists() and file_2.ensure_file_exists() \
                and skip_if_exists:
            return file_1, file_2

        with ResultsPlot(dest_file=file_1, sup_title=the_title,
                         state_dims=self.state_dims,
                         plot_indices=self.plot_examples) as plt, \
                ResultsLog(self.state_dims, file_2) as log:
            multi_run_ode(
                self.test_starting_states,
                self.training_starting_states,
                (plt.collector, log.collector),
                self.equations,
                controller, parameters, self.control_dims,
                self.test_steps, self.training_steps)

        return file_1, file_2

    def describe_system_without_control(self, dest_dir: str,
                                        skip_if_exists: bool = True) \
            -> tuple[Path, Path]:
        """
        Describe the performance of a given system of system.

        :param dest_dir: the destination directory
        :param skip_if_exists: if the file already exists
        :returns: the paths of the generated files
        """
        return self.describe_system("without control", _dummy_ctrl, 0.0,
                                    f"{self.name}_without_control", dest_dir,
                                    skip_if_exists)



"""
A primitive integrator for systems of ordinary differential equations.

Many dynamic systems can be modeled as systems of ordinary differential
equations that govern their progress over time. Trying to find out in
which state such systems are at a given point in time means to integrate
these equations until that point in time (starting from a starting state).
Now there are several methods for doing such a thing, including, e.g.,
:func:`scipy.integrate.odeint`.

What we want to play around with, however, is synthesizing controllers.
In this case, the differential equations also merge the output of the
controller with the current state. If the controller behaves inappropriately,
this may make the system diverge, i.e., some of its state variables go to
infinity over time or sometimes rather quickly.

Using ODE integrators that compute the system state at pre-defined time steps
is thus cumbersome, as the system may have already exploded at these goal
times. Therefore, we cooked our own, simple ODE integrator. Our integrator
dynamically chooses the length of the time step based on the output of the
differential equations. If the output is large, then we will automatically use
shorter time slices. A large differential applied for a small time slice still
leads to a reasonably-large change in the system state. More specifically, the
system cannot diverge to infinity: The larger the differential step, the
shorter the time slice we use, so everything balances out.

Thus, we can simulate a well-behaved system over a long time and an
ill-behaved system for a shorter time period. Neither system will diverge.
"""

from math import fsum, sqrt
from typing import Callable, Final, Iterable, TypeVar

import numba  # type: ignore
import numpy as np


@numba.njit(numba.float64(numba.float64[:], numba.float64[:],
                          numba.float64[:], numba.float64[:], numba.float64),
            cache=True, inline="always", fastmath=False, boundscheck=False)
def __diff_update(cur_diff: np.ndarray, cur_row: np.ndarray,
                  last_diff: np.ndarray, next_row: np.ndarray,
                  max_dist: float) -> float:
    """
    Add the current differential to the current state to get the next state.

    The input of this function is the current differential vector `cur_diff`,
    the current row of the ode matrix `cur_row`, the last differential vector
    `last_diff`, the next row of the ode matrix `next_row`, and the maximum
    permitted step length `max_dist`.
    In an ideal world, we would set the weight value of the current row, i.e.,
    `cur_row[-1]` to `1.0` and set `next_row[0:s] = cur_row[0:s] + cur_diff`,
    where `s` be the dimension of the state. In other words, we would just add
    the differential vector to the current state and get the next state.
    However, we do not live in an ideal world. The absolute value
    `z = |cur_diff| = np.sqrt(np.square(cur_diff).sum())`
    may be too large. This can have two reasons:

    1. Maybe the differential equation system diverges, i.e., the system
       moves towards infinity.
    2. Maybe we are close to some state towards the system may converge, but
       the step length does not become small enough to actually converge and
       the system jumps around.

    So we put several mechanisms in place to limit the step lengths:

    1. We never use the full step length `z` but pre-scale it to
       `log(1 + z) * 0.6180339887498948`.
    2. We limit the step length to the 1/128th of the distance of the current
       system state to the center of coordinates.
    3. We multiply the step length with a factor that exponentially decreases
       with the angle between the current and the last differential.
    4. We impose a hard limit on the step length based on the dimensionality
       of the system.

    All of this leads to a smaller step length `t <= z`. The weight of the
    step is then `t / z` (if `z > 0`, otherwise `1`)

    :param cur_diff: the differential vector
    :param cur_row: the current row
    :param last_diff: the last differential vector
    :param next_row: the state to update
    :param max_dist: the maximum distance to move
    :returns: the current weight
    :returns: the length of the time slice

    >>> c_row = np.array([0.0, 0.0, 0.0, 0.0], float)
    >>> c_diff = np.array([1.0, 1.0, 1.0], float)
    >>> l_diff = np.zeros_like(c_diff)
    >>> n_row = np.zeros_like(c_row)
    >>> _ = __diff_update(c_diff, c_row, l_diff, n_row, 100.0)  # log-limit
    >>> print(";".join(f"{s:.10f}" for s in n_row))
    0.3586249472;0.3586249472;0.3586249472;0.0000000000
    >>> print(f"{c_row[-1]:.10f}")
    0.3586249472
    >>> print((np.log1p(np.sqrt(3.0)) * 0.6180339887498948) / np.sqrt(3.0))
    0.3586249472058058
    >>> _ = __diff_update(c_diff, c_row, l_diff, n_row, 0.05)  # hard limit
    >>> print(";".join(f"{s:.10f}" for s in n_row))
    0.0288675135;0.0288675135;0.0288675135;0.0000000000
    >>> print(f"{c_row[-1]:.10f}")
    0.0288675135
    >>> print(0.05 / np.sqrt(3.0))
    0.02886751345948129
    >>> l_diff = np.array([-1.0, -1.0, -1.0], float)
    >>> _ = __diff_update(c_diff, c_row, l_diff, n_row, 100.0)  # angle-limit
    >>> print(";".join(f"{s:.10f}" for s in n_row))
    0.0000289409;0.0000289409;0.0000289409;0.0000000000
    >>> print(f"{c_row[-1]:.10f}")
    0.0000289409
    >>> a = (np.log1p(np.sqrt(3.0)) * 0.6180339887498948
    ...      * np.exp(-3 * np.pi)) / np.sqrt(3.0)
    >>> print(f"{a:.10f}")
    0.0000289409
    >>> l_diff = np.array([-1.0, -1.0, 1.0], float)
    >>> _ = __diff_update(c_diff, c_row, l_diff, n_row, 100.0)  # angle-limit
    >>> print(";".join(f"{s:.10f}" for s in n_row))
    0.0011622728;0.0011622728;0.0011622728;0.0000000000
    >>> print(f"{c_row[-1]:.10f}")
    0.0011622728
    >>> a = (np.log1p(np.sqrt(3.0)) * 0.6180339887498948
    ...      * np.exp(-3 * np.arccos(-1 / 3))) / np.sqrt(3.0)
    >>> print(f"{a:.10f}")
    0.0011622728
    >>> c_row = np.array([0.1, 0.1, 0.1, 0.0], float)
    >>> l_diff = np.zeros_like(c_diff)
    >>> _ = __diff_update(c_diff, c_row, l_diff, n_row, 100.0)  # center limit
    >>> print(";".join(f"{s:.10f}" for s in n_row))
    0.1007812500;0.1007812500;0.1007812500;0.0000000000
    >>> print(f"{c_row[-1]:.10f}")
    0.0007812500
    >>> print((np.sqrt(0.03) / 128) / np.sqrt(3))
    0.00078125
    """
# Compute the square sum of the current differential.
    dsum: float = 0.0
    c: float = 0.0
    failure: bool = False
    for dx in cur_diff:
        if -1e100 < dx < 1e100:
            y = (dx * dx) - c
            t = dsum + y
            c = (t - dsum) - y
            dsum = t
        else:
            failure = True
            break

# If the differential grew out of bounds, we just add a logarithmic step.
    if failure or (dsum > 1e100):
        for i, d in enumerate(cur_diff):
            next_row[i] = cur_row[i] + (
                -np.log1p(-d) if d < 0.0 else np.log1p(d))
        cur_row[-1] = 1.0
        return 1.0

    dist: Final[float] = np.sqrt(dsum)
    dim: Final[int] = len(cur_diff)
    if dist <= 0.0:  # If the step length if 0, we can quit directly.
        next_row[0:dim] = cur_row[0:dim]
        next_row[-1] = 1.0
        return 1.0

# the first adjustment of the step length is always log(1+dist) / golden ratio
    use_dist: float = np.log1p(dist) * 0.6180339887498948

# now compute the angle-based limit
# The angle between two vectors A and B is arccos(A*B / (|A| * |B|)).
    p_dsum: float = 0.0  # this will be used to compute |B|
    prod: float = 0.0  # this is used to compute A*B
    c2: float = 0.0
    for i, dd in enumerate(last_diff):
        if -1e100 < dd < 1e100:
            y = (dd * dd) - c
            t = p_dsum + y
            c = (t - p_dsum) - y
            p_dsum = t

            y = (dd * cur_diff[i]) - c2
            t = prod + y
            c2 = (t - prod) - y
            prod = t
        else:
            failure = True
            break

    if (not failure) and (p_dsum > 0.0):
        p_dsum = np.sqrt(p_dsum) * dist
        if p_dsum > 0.0:
            prod /= p_dsum
            if prod < 1.0:
                use_dist *= np.exp(
                    -3.0 * (np.arccos(prod) if prod > -1.0 else np.pi))

# now limit the step length to 1/128th of the distance to the center
    center_dist: float = 0.0
    c = 0.0
    for i in range(dim):
        dd = cur_row[i]
        if -1e100 < dd < 1e100:
            y = (dd * dd) - c
            t = center_dist + y
            c = (t - center_dist) - y
            center_dist = t
        else:
            center_dist = max_dist
            break
    center_dist = 0.0078125 * np.sqrt(center_dist)
    if 0.0 < center_dist < use_dist:
        use_dist = center_dist

# now apply the hard, dimension-based limit
    if max_dist < use_dist:
        use_dist = max_dist

# store the weight in the current row and update the next row
    cur_row[-1] = weight = use_dist / dist
    next_row[0:dim] = cur_row[0:dim] + (weight * cur_diff)
    return weight


#: the type variable for ODE controller parameters
T = TypeVar("T")


def run_ode(starting_state: np.ndarray,
            equations: Callable[[
                np.ndarray, float, np.ndarray, np.ndarray], None],
            controller: Callable[[np.ndarray, float, T, np.ndarray], None],
            parameters: T, controller_dim: int = 1,
            steps: int = 5000) -> np.ndarray:
    """
    Simulate a set of controlled differential system.

    The system begins in the starting state stored in the vector
    `starting_state`. In each time step, first, the `controller` is invoked.
    It receives the current system state vector as input, the current time
    `t`, its parameters (`parameters`), and an output array to store its
    computed control values into. This array has dimension `controller_dim`,
    which usually is `1`. Then, the function `system` will be called and
    receives the current state vector, the time step `t`, and the controller
    output as input, as well as an output array to store the result of the
    differential into. This output array has the same dimension as the state
    vector.

    The system will be simulated for exactly `steps` time steps. The length of
    the time steps will automatically be determined based on a quick
    preliminary simulation of the system and controller. A maximum time
    index `max_time` can be provided. The system will automatically determine
    the right length of the time steps.

    As result, this function returns a tuple of three components: First, the
    matrix `steps` rows, where each row corresponds to one system state
    vector. The first row will be identical tp `starting_state`. The second
    element of the return tuple is the time vector, holding `steps` strictly
    increasing time value. The third and last return component is the control
    matrix. It has one row for each time step, i.e., `steps` in total, and
    each row corresponds to the controller output at that time step. If
    `controller_dim==1`, this third tuple element is a plain vector and not a
    matrix.

    The last time slice (at index `-1,-1`) will always be set to zero.

    :param starting_state: the starting
    :param equations: the differential system
    :param controller: the controller function
    :param parameters: the controller parameters
    :param controller_dim: the dimension of the controller result
    :param steps: the number of steps to simulate
    :returns: a matrix where each row represents a point in time, composed of
        the current state, the controller output, and the length of the time
        slice

    If we simulate the flight of a projectile with our ODE execution, then
    both the flight time as well as the flight length are about 0.12% off from
    what the mathematical solution of the flight system prescribe. That's
    actually not bad for a crude and fast integration method...

    >>> v = 100.0
    >>> angle = np.deg2rad(45.0)
    >>> v_x = v * np.cos(angle)
    >>> print(f"{v_x:.10f}")
    70.7106781187
    >>> v_y = v * np.sin(angle)
    >>> print(f"{v_y:.10f}")
    70.7106781187
    >>> def projectile(position, ttime, ctrl, out):
    ...     out[0] = 70.71067811865474
    ...     out[1] = 70.71067811865474 - ttime * 9.80665
    >>> param = 0.0   # ignore
    >>> def contrl(position, ttime, params, dest):
    ...     dest[0] = 0.0  #  controller that does nothing
    >>> strt = np.array([0.0, 1.0])
    >>> ode = run_ode(strt, projectile, contrl, param, 1, 10000)
    >>> time_of_flight = 2 * v_y / 9.80665
    >>> print(f"{time_of_flight:.10f}")
    14.4209649817
    >>> travel_distance_x = time_of_flight * v_x
    >>> print(f"{travel_distance_x:.10f}")
    1019.7162129779
    >>> idx = np.argwhere(ode[:, 1] <= 0.0)[0][0]
    >>> print(idx)
    9631
    >>> print(f"{ode[idx - 1, 0]:.10f}")
    1020.8026525593
    >>> print(f"{ode[idx, 0]:.10f}")
    1020.8909467853
    >>> print(f"{fsum(ode[0:idx, -1]):.10f}")
    14.4375782265
    >>> print(f"{fsum(ode[0:idx + 1, -1]):.10f}")
    14.4388267873
    >>> print(ode[-1, -1])
    0.0
    """
    state_dim: Final[int] = len(starting_state)
    result: Final[np.ndarray] = np.zeros((
        steps, state_dim + controller_dim + 1))
    max_dist: Final[float] = sqrt(
        (state_dim - 1) / 64.0)  # the maximum distance

    point: np.ndarray = np.empty(state_dim)
    cur_diff: np.ndarray = np.empty(state_dim)
    last_diff: np.ndarray = np.zeros(state_dim)
    c_out: Final[np.ndarray] = np.empty(controller_dim)
    time: float = 0.0
    c: float = 0.0
    c_end_dim: Final[int] = state_dim + controller_dim

    next_row: np.ndarray = result[0]
    next_row[0:state_dim] = starting_state

    # compute the first row
    for i in range(1, steps):
        cur_row = next_row
        next_row = result[i]
        point[:] = cur_row[0:state_dim]  # get point part of state
        controller(point, time, parameters, c_out)  # invoke controller
        equations(point, time, c_out, cur_diff)  # invoke equations

# invoke the differential update and update time as well
        y = __diff_update(cur_diff, cur_row, last_diff,
                          next_row, max_dist) - c
        t = time + y
        c = (t - time) - y
        time = t

        cur_row[state_dim:c_end_dim] = c_out  # store controller output
        last_diff, cur_diff = cur_diff, last_diff

    point[:] = next_row[0:state_dim]
    controller(point, time, parameters, c_out)
    next_row[state_dim:c_end_dim] = c_out
    return result


@numba.njit(cache=True, inline="always", fastmath=False, boundscheck=False)
def __j_from_ode_compute(ode: np.ndarray, state_dim: int,
                         dest: np.ndarray) -> None:
    """
    Prepare the input array for the figure of merit computation.

    The `ode` matrix contains one row for each time step.
    The row is composed of the current state, the current controller output,
    and the length of the time slice.
    However, all systems of a given simulation start in the same initial
    state, so it makes little sense to include this initial state into the
    figure of merit computation. It does make sense to include the control
    output at that moment, though, because it contributes to the next state.
    It also makes no sense to count in the last row of the ODE computation,
    because this is the final system state whose weight will not actually
    be computed.

    :param ode: the output array from the ODE simulation
    :param state_dim: the state dimension
    :param dest: the destination array

    >>> od = np.array([[1, 2, 3, 4, 0.1],
    ...                [5, 6, 7, 8, 0.2],
    ...                [9, 6, 4, 3, 0.3],
    ...                [7, 4, 2, 1, 0.4]])
    >>> sta_dim = 3
    >>> dst = np.empty((od.shape[0] - 1) * (od.shape[1] - 1) - sta_dim)
    >>> __j_from_ode_compute(od, sta_dim, dst)
    >>> print(dst)
    [24.3  10.8   4.8   0.27  5.    7.2   9.8   1.28  0.16]
    >>> rs = np.array([9 * 9 * 0.3, 6 * 6 * 0.3, 4 * 4 * 0.3, 3 * 3 * 0.03,
    ...                5 * 5 * 0.2, 6 * 6 * 0.2, 7 * 7 * 0.2, 8 * 8 * 0.02,
    ...                4 * 4 * 0.01])
    >>> print(rs)
    [24.3  10.8   4.8   0.27  5.    7.2   9.8   1.28  0.16]
    """
    index: int = len(dest)
    start: Final[int] = ode.shape[1] - 2

    inner: int = start
    row = ode[0]
    weight_01: float = row[-1] * 0.1
    while inner >= state_dim:
        v = row[inner]
        inner -= 1
        index -= 1
        dest[index] = (v * v) * weight_01 if -1e100 < v < 1e100 else 1e100
    for row in ode[1:-1]:
        inner = start
        weight: float = row[-1]
        weight_01 = weight * 0.1
        while inner >= state_dim:
            v = row[inner]
            inner -= 1
            index -= 1
            dest[index] = (v * v) * weight_01 if -1e100 < v < 1e100 else 1e100
        while inner >= 0:
            v = row[inner]
            inner -= 1
            index -= 1
            dest[index] = (v * v) * weight if -1e100 < v < 1e100 else 1e100


def t_from_ode(ode: np.ndarray) -> float:
    """
    Get the time sum from an ODE solution.

    The total time that we simulate a system depends on the behavior of the
    system. If the system makes small steps (has small differential values)
    and does not change its direction, it can be simulated for a long time.
    If the system diverges, e.g., if some state variables would go to
    infinity, or if it often changes direction or has some larger steps, then
    the total simulation time shortens automatically. This prevents the system
    from actually diverging.

    :param ode: the ODE solution, as return from :func:`run_ode`.
    :return: the time sum

    >>> od = np.array([[1, 2, 3, 4, 0.1],
    ...                [5, 6, 7, 8, 0.2],
    ...                [9, 6, 4, 3, 0.3],
    ...                [7, 4, 2, 1, 0.4]])
    >>> print(t_from_ode(od))
    0.6
    """
    return fsum(ode[0:-1, -1])


def j_from_ode(ode: np.ndarray, state_dim: int) -> float:
    """
    Compute the original figure of merit from an ODE array.

    The figure of merit is the sum of state variable squares plus 0.1 times
    the control variable squares. We disregard the state variable values of
    the starting states (because they are the same for all controllers on
    a given training case and because the control cannot influence them) and
    we also disregard the final state and final controller output (as there is
    no time slice associated with them, i.e., we only "arrive" in them but
    basically spent 0 time in them in our simulation).

    :param ode: the array returned by the ODE function, i.e., :func:`run_ode`
    :param state_dim: the state dimension
    :return: the figure of merit

    >>> od = np.array([[1, 2, 3, 4, 0.1],
    ...                [5, 6, 7, 8, 0.2],
    ...                [9, 6, 4, 3, 0.3],
    ...                [7, 4, 2, 1, 0.4]])
    >>> sta_dim = 3
    >>> print(f"{j_from_ode(od, sta_dim):.10f}")
    106.0166666667
    >>> print((24.3 + 10.8 + 4.8 + 0.27 + 5. + 7.2 + 9.8 + 1.28 + 0.16) / 0.6)
    106.01666666666667
    """
    dest: Final[np.ndarray] = np.empty(
        (ode.shape[0] - 1) * (ode.shape[1] - 1) - state_dim)
    __j_from_ode_compute(ode, state_dim, dest)
    return fsum(dest) / t_from_ode(ode)


def multi_run_ode(
        test_starting_states: Iterable[np.ndarray],
        training_starting_states: Iterable[np.ndarray],
        collector: Callable[[int, np.ndarray, float, float], None] | Iterable[
            Callable[[int, np.ndarray, float, float], None]],
        equations: Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None],
        controller: Callable[[np.ndarray, float, T, np.ndarray], None],
        parameters: T, controller_dim: int = 1,
        test_steps: int = 5000, training_steps: int = 5000) -> None:
    """
    Invoke :func:`run_ode` multiple times and pass the result to `collector`.

    This function allows us to perform multiple runs of the differential
    equation simulator, using different starting points. It also allows us to
    distinguish training and test points and to assign them different numbers
    of steps. For each of them, :func:`run_ode` will be applied and the
    returned matrix is passed to the `collector` function.

    :param test_starting_states: the iterable of test starting states
    :param training_starting_states: the iterable of training starting states
    :param collector: the destination to receive the results
    :param equations: the differential system
    :param controller: the controller function
    :param parameters: the controller parameters
    :param controller_dim: the dimension of the controller result
    :param test_steps: the number of test steps to simulate
    :param training_steps: the number of training steps to simulate
    """
    if not isinstance(collector, Iterable):
        collector = (collector, )
    index: int = 0
    for sp in test_starting_states:
        ode = run_ode(sp, equations, controller, parameters, controller_dim,
                      test_steps)
        for c in collector:
            c(index, ode, j_from_ode(ode, len(sp)), t_from_ode(ode))
        index += 1
    for sp in training_starting_states:
        ode = run_ode(sp, equations, controller, parameters, controller_dim,
                      training_steps)
        for c in collector:
            c(index, ode, j_from_ode(ode, len(sp)), t_from_ode(ode))
        index += 1



"""Poor man's Artificial Neural Networks."""

from typing import Final, Iterable

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_2d_1o_1(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a man's ANN for 2d-spaces with one layer containing 1 node.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    out[0] = params[0] * np.arctan(params[1] + (params[2] * state[0])
                                   + (params[3] * state[1]))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_3d_1o_1(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 3d-spaces with one layer containing 1 node.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    out[0] = params[0] * np.arctan(
        params[1] + (params[2] * state[0]) + (params[3] * state[1])
        + (params[4] * state[2]))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_2d_1o_2(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 2d-spaces with one layer containing 2 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1))
    hl_1_2 = np.arctan(params[3] + (params[4] * s0) + (params[5] * s1))
    out[0] = (params[6] * hl_1_1) + (params[7] * hl_1_2)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_3d_1o_2(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 3d-spaces with one layer containing 2 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1)
                       + (params[3] * s2))
    hl_1_2 = np.arctan(params[4] + (params[5] * s0) + (params[6] * s1)
                       + (params[7] * s2))
    out[0] = (params[8] * hl_1_1) + (params[9] * hl_1_2)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_2d_1o_3(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 2d-spaces with one layer containing 3 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1))
    hl_1_2 = np.arctan(params[3] + (params[4] * s0) + (params[5] * s1))
    hl_1_3 = np.arctan(params[6] + (params[7] * s0) + (params[8] * s1))
    out[0] = (params[9] * hl_1_1) + (params[10] * hl_1_2) \
        + (params[11] * hl_1_3)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_3d_1o_3(state: np.ndarray, _: float,
                  params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 3d-spaces with one layer containing 3 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1)
                       + (params[3] * s2))
    hl_1_2 = np.arctan(params[4] + (params[5] * s0) + (params[6] * s1)
                       + (params[7] * s2))
    hl_1_3 = np.arctan(params[8] + (params[9] * s0) + (params[10] * s1)
                       + (params[11] * s2))
    out[0] = (params[12] * hl_1_1) + (params[13] * hl_1_2) \
        + (params[14] * hl_1_3)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_2d_1o_2_2(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 2d-spaces with 2 layers with 2 nodes each.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1))
    hl_1_2 = np.arctan(params[3] + (params[4] * s0) + (params[5] * s1))
    hl_2_1 = np.arctan(params[6] + (params[7] * hl_1_1)
                       + (params[8] * hl_1_2))
    hl_2_2 = np.arctan(params[9] + (params[10] * hl_1_1)
                       + (params[11] * hl_1_2))
    out[0] = (params[12] * hl_2_1) + (params[13] * hl_2_2)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_3d_1o_2_2(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 3d-spaces with 2 layers with 2 nodes each.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1)
                       + (params[3] * s2))
    hl_1_2 = np.arctan(params[4] + (params[5] * s0) + (params[6] * s1)
                       + (params[7] * s2))
    hl_2_1 = np.arctan(params[8] + (params[9] * hl_1_1)
                       + (params[10] * hl_1_2))
    hl_2_2 = np.arctan(params[11] + (params[12] * hl_1_1)
                       + (params[13] * hl_1_2))
    out[0] = (params[14] * hl_2_1) + (params[15] * hl_2_2)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_2d_1o_3_2(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 2d-spaces with 2 layers with 3 and 2 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1))
    hl_1_2 = np.arctan(params[3] + (params[4] * s0) + (params[5] * s1))
    hl_1_3 = np.arctan(params[6] + (params[7] * s0) + (params[8] * s1))
    hl_2_1 = np.arctan(params[9] + (params[10] * hl_1_1)
                       + (params[11] * hl_1_2) + (params[12] * hl_1_3))
    hl_2_2 = np.arctan(params[13] + (params[14] * hl_1_1)
                       + (params[15] * hl_1_2) + (params[16] * hl_1_2))
    out[0] = (params[17] * hl_2_1) + (params[18] * hl_2_2)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __ann_3d_1o_3_2(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a poor man's ANN for 3d-spaces with 2 layers with 3 and 2 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    hl_1_1 = np.arctan(params[0] + (params[1] * s0) + (params[2] * s1)
                       + (params[3] * s2))
    hl_1_2 = np.arctan(params[4] + (params[5] * s0) + (params[6] * s1)
                       + (params[7] * s2))
    hl_1_3 = np.arctan(params[8] + (params[9] * s0) + (params[10] * s1)
                       + (params[11] * s2))
    hl_2_1 = np.arctan(params[12] + (params[13] * hl_1_1)
                       + (params[14] * hl_1_2) + (params[15] * hl_1_3))
    hl_2_2 = np.arctan(params[16] + (params[17] * hl_1_1)
                       + (params[18] * hl_1_2) + (params[19] * hl_1_2))
    out[0] = (params[20] * hl_2_1) + (params[21] * hl_2_2) \
        + (params[22] * hl_2_2)


def anns(system: System) -> Iterable[Controller]:
    """
    Create poor man's ANNs fitting to a given system.

    Based on the dimensionality of the state space, we generate a set of ANNs
    with different numbers of layers and neurons. The weights of the neurons
    can then directly be optimized by a numerical optimization algorithm.
    This is, of course, probably much less efficient that doing some proper
    learning like back-propagation. However, it allows us to easily plug the
    ANNs into the same optimization routines as other controllers.

    :param system: the equations object
    :return: the ANNs
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    if system.state_dims == 2:
        return (Controller("ann_1", 2, 1, 4, __ann_2d_1o_1),
                Controller("ann_2", 2, 1, 8, __ann_2d_1o_2),
                Controller("ann_3", 2, 1, 12, __ann_2d_1o_3),
                Controller("ann_2_2", 2, 1, 14, __ann_2d_1o_2_2),
                Controller("ann_3_2", 2, 1, 19, __ann_2d_1o_3_2))
    if system.state_dims == 3:
        return (Controller("ann_1", 3, 1, 5, __ann_3d_1o_1),
                Controller("ann_2", 3, 1, 10, __ann_3d_1o_2),
                Controller("ann_3", 3, 1, 15, __ann_3d_1o_3),
                Controller("ann_2_2", 3, 1, 16, __ann_3d_1o_2_2),
                Controller("ann_3_2", 3, 1, 23, __ann_3d_1o_3_2))
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
An objective function for the dynamic control problem.

The dynamic control problem means that our system starts in a given state
and we try to move it to a stable state by using control effort. Controllers
are trained over several training states, for each of which we can compute a
figure of merit. This objective function here just averages over these figures
of merit. Maybe :mod:`~moptipyapps.dynamic_control.objective_le`, which tries
to smooth out the impact of bad starting states leads to better results,
though.
"""

from typing import Callable, Final

import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error

from moptipyapps.dynamic_control.instance import Instance
from moptipyapps.dynamic_control.ode import j_from_ode, run_ode
from moptipyapps.shared import SCOPE_INSTANCE


class FigureOfMerit(Objective):
    """Compute the figure of merit for the given instance."""

    def __init__(self, instance: Instance) -> None:
        """
        Create the figure-of-merit objective of the dynamic control problem.

        :param instance: the instance
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the dynamic control instance
        self.instance: Final[Instance] = instance

        #: the simulation steps
        self.__steps: Final[int] = instance.system.training_steps
        #: the training starting states
        self.__training: Final[np.ndarray] = \
            instance.system.training_starting_states
        #: the results
        self.__results: Final[np.ndarray] = np.empty(
            len(self.__training), float)
        #: the equations
        self.__equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] = \
            instance.system.equations
        #: the controller
        self.__controller: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] = \
            instance.controller.controller
        #: the controller dimension
        self.__controller_dim: Final[int] = instance.controller.control_dims

    def __str__(self) -> str:
        """
        Get the name of this objective function.

        :return: `figureOfMerit`
        """
        return "figureOfMerit"

    def evaluate(self, x: np.ndarray) -> float:
        """
        Evaluate the parameterization of a controller.

        :param x: the controller parameters
        :return: the figure of merit
        """
        steps: Final[int] = self.__steps
        training: Final[np.ndarray] = self.__training
        results: Final[np.ndarray] = self.__results
        equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] \
            = self.__equations
        controller: Final[
            Callable[[np.ndarray, float, np.ndarray, np.ndarray], None]] \
            = self.__controller
        controller_dim: Final[int] = self.__controller_dim

        for i, start in enumerate(training):
            results[i] = j_from_ode(run_ode(
                start, equations, controller, x, controller_dim, steps),
                len(start))
        return results.mean()

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as scope:
            self.instance.log_parameters_to(scope)

    def lower_bound(self) -> float:
        """
        Get the lower bound of the figure of merit, which is 0.

        :returns: 0.0
        """
        return 0.0



"""
Applications of Metaheuristic Optimization in Python.

Currently, the following applications are implemented:

- :mod:`~moptipyapps.binpacking2d` provides methods to solve two-dimensional
  bin packing instances.

The following additional tools are implemented:

- :mod:`~moptipyapps.tests` offers unit tests to try out optimization
  algorithms and other instances of :mod:`~moptipy.api.component` on the
  different problems that are provided above.
- :mod:`~moptipyapps.shared` offers shared constants and tools.
"""


def __setup() -> None:
    """Add `moptipyapps` to the dependencies unless called from pytest."""
    import inspect  # noqa # pylint: disable=C0415
    import moptipy.utils.sys_info  # noqa # pylint: disable=C0415

    for s in inspect.stack():
        if "pytest" in s.filename:
            return
    moptipy.utils.sys_info.add_dependency("moptipyapps")


__setup()
del __setup



"""
Examples of dynamic control problems.

Here we have examples for dynamic control problems that are systems whose
state is governed by differential system which can be influenced by the
output of controllers.
Such systems can serve as primitive testbeds of actual dynamic control
scenarios such as flight control or other fluid dynamic systems. Different
from such more complicated systems, they are much faster and easier to
simulate, though. So we can play with them much more easily and quickly.

The initial starting point of the work here were conversations with
Prof. Dr. Bernd NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute
of Technology in Shenzhen, China (哈尔滨工业大学(深圳)) as well as the
following two MSc theses and book:

1. Yuxiang LI (李宇翔). Jet Mixing Enhancement using Deep Reinforcement
   Learning (基于深度强化学习的射流混合增强控制). MSc Thesis. Harbin Institute
   of Technology in Shenzhen, China (哈尔滨工业大学(深圳)).
   January 2023.
2. Wei SUN (孙伟). Wake Control of 1-2-3 Fluidic Pinball using Deep
   Reinforcement Learning (基于深度强化学习方法的 1-2-3 流体弹球尾流控制).
   MSc Thesis. Harbin Institute of Technology in Shenzhen, China
   (哈尔滨工业大学(深圳)). January 2023.
3. Guy Yoslan CORNEJO MACEDA, François LUSSEYRAN, and Bernd R. NOACK.
   xMLC: A Toolkit for Machine Learning Control, First Edition.
   Machine Learning Tools in Fluid Mechanics, Vol 2.
   Shenzhen & Paris; Universitätsbibliothek der Technischen Universität
   Braunschweig. 2022 https://doi.org/10.24355/dbbs.084-202208220937-0
"""



"""A set of pre-defined controllers."""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __cornejo_maceda(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a parameterized version of Cornejo Maceda's Law.

    See page 16, chapter 3.3.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    z = np.tanh(state[0] - state[1])
    b = params[0]
    z = np.tanh(1.0 if b == 0 else z / b)
    b = params[1]
    z = np.tanh(1.0 if b == 0 else z / b)
    b = params[2]
    out[0] = np.tanh(1.0 if b == 0 else z / b)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __table_3_1_ga(state: np.ndarray, _: float,
                   params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute the law evolved by genetic algorithm in Table 3-1.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    out[0] = state[0] * params[0] + state[1] * params[1]


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __table_3_1_lgpc(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute the law evolved by LGPC algorithm in Table 3-1.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    a: Final[float] = state[0] * params[0] + params[1]
    out[0] = params[2] * np.sin((params[3] / a) if (a != 0.0) else 1.0)


def predefined(system: System) -> tuple[Controller, ...]:
    """
    Create a set of pre-defined controllers for the given equations object.

    :param system: the equations object
    :return: the linear controller
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    if system.state_dims == 2:
        return (Controller("cornejo_maceda", 2, 1, 3, __cornejo_maceda), )
    if system.state_dims == 3:
        return (Controller("table_3_1_ga", 3, 1, 2, __table_3_1_ga),
                Controller("table_3_1_lgpc", 3, 1, 4, __table_3_1_lgpc))
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""An extended end result statistics record to represent packings."""
import os.path
from dataclasses import dataclass
from math import isfinite
from typing import Callable, Final, Iterable, Mapping, cast

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.evaluation.base import (
    KEY_N,
)
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import (
    KEY_BEST_F_SCALED,
    KEY_ERT_FES,
    KEY_ERT_TIME_MILLIS,
    KEY_N_SUCCESS,
    KEY_SUCCESS_FES,
    KEY_SUCCESS_TIME_MILLIS,
    EndStatistics,
)
from moptipy.evaluation.statistics import (
    EMPTY_CSV_ROW,
    Statistics,
)
from moptipy.utils.console import logger
from moptipy.utils.logger import CSV_SEPARATOR
from moptipy.utils.path import Path
from moptipy.utils.strings import (
    num_to_str,
)
from moptipy.utils.types import check_int_range, immutable_mapping, type_error

from moptipyapps.binpacking2d.bin_count import BIN_COUNT_NAME
from moptipyapps.binpacking2d.packing_result import (
    _OBJECTIVE_LOWER,
    _OBJECTIVE_UPPER,
    KEY_BIN_HEIGHT,
    KEY_BIN_WIDTH,
    KEY_N_DIFFERENT_ITEMS,
    KEY_N_ITEMS,
    KEY_USED_OBJECTIVE,
    PackingResult,
)


@dataclass(frozen=True, init=False, order=True)
class PackingStatistics:
    """
    An end statistics record of one run of one algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: the original end statistics record
    end_statistics: EndStatistics
    #: the number of items in the instance
    n_items: int
    #: the number of different items in the instance
    n_different_items: int
    #: the bin width
    bin_width: int
    #: the bin heigth
    bin_height: int
    #: the objective values evaluated after the optimization
    objectives: Mapping[str, Statistics]
    #: the objective function used for optimization
    used_objective: str
    #: the bounds for the objective values (append ".lowerBound" and
    #: ".upperBound" to all objective function names)
    objective_bounds: Mapping[str, int]
    #: the bounds for the minimum number of bins of the instance
    bin_bounds: Mapping[str, int]

    def __init__(self,
                 n_items: int,
                 n_different_items: int,
                 bin_width: int,
                 bin_height: int,
                 end_statistics: EndStatistics,
                 objectives: Mapping[str, Statistics],
                 used_objective: str,
                 objective_bounds: Mapping[str, int | float],
                 bin_bounds: Mapping[str, int]):
        """
        Create a consistent instance of :class:`PackingStatistics`.

        :param end_statistics: the end statistics
        :param n_items: the number of items
        :param n_different_items: the number of different items
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param objectives: the objective values computed after the
            optimization
        :param used_objective: the objective function used
        :param bin_bounds: the different bounds for the number of bins
        :param objective_bounds: the bounds for the objective functions
        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__()
        if not isinstance(end_statistics, EndStatistics):
            raise type_error(end_statistics, "end_statistics", EndResult)
        if not isinstance(used_objective, str):
            raise type_error(used_objective, "used_objective", str)
        if end_statistics.best_f != objectives[used_objective]:
            raise ValueError(
                f"end_statistics.best_f={end_statistics.best_f}, but objecti"
                f"ves[{used_objective!r}]={objectives[used_objective]}.")
        if not isinstance(objectives, Mapping):
            raise type_error(objectives, "objectives", Mapping)
        if not isinstance(objective_bounds, Mapping):
            raise type_error(objective_bounds, "objective_bounds", Mapping)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        if len(objective_bounds) != (2 * len(objectives)):
            raise ValueError(f"it is required that there is a lower and an "
                             f"upper bound for each of the {len(objectives)} "
                             f"functions, but we got {len(objective_bounds)} "
                             f"bounds, objectives={objectives}, "
                             f"objective_bounds={objective_bounds}.")

        for name, value in objectives.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of evaluation[{name!r}]={value!r}", str)
            if not isinstance(value, Statistics):
                raise type_error(
                    value, f"value of evaluation[{name!r}]={value!r}",
                    Statistics)
            lower = objective_bounds[f"{name}{_OBJECTIVE_LOWER}"]
            if not isfinite(lower):
                raise ValueError(f"{name}{_OBJECTIVE_LOWER}=={lower}.")
            upper = objective_bounds[f"{name}{_OBJECTIVE_UPPER}"]
            if not (lower <= value.minimum <= value.maximum <= upper):
                raise ValueError(
                    f"it is required that {name}{_OBJECTIVE_LOWER}<={name}."
                    f"min<={name}.max<={name}{_OBJECTIVE_UPPER}, but got "
                    f"{lower}, {value.minimum}, {value.maximum} and {upper}.")

        bins: Final[Statistics | None] = cast(
            Statistics, objectives[BIN_COUNT_NAME]) \
            if BIN_COUNT_NAME in objectives else None
        for name, value2 in bin_bounds.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of bounds[{name!r}]={value2!r}", str)
            check_int_range(value2, f"bounds[{name!r}]", 1, 1_000_000_000)
            if (bins is not None) and (bins.minimum < value2):
                raise ValueError(
                    f"number of bins={bins} is inconsistent with "
                    f"bound {name!r}={value2}.")

        object.__setattr__(self, "end_statistics", end_statistics)
        object.__setattr__(self, "objectives", immutable_mapping(objectives))
        object.__setattr__(self, "used_objective", used_objective)
        object.__setattr__(self, "objective_bounds",
                           immutable_mapping(objective_bounds))
        object.__setattr__(self, "bin_bounds", immutable_mapping(bin_bounds))
        object.__setattr__(self, "n_different_items", check_int_range(
            n_different_items, "n_different_items", 1, 1_000_000_000_000))
        object.__setattr__(self, "n_items", check_int_range(
            n_items, "n_items", n_different_items, 1_000_000_000_000))
        object.__setattr__(self, "bin_width", check_int_range(
            bin_width, "bin_width", 1, 1_000_000_000_000))
        object.__setattr__(self, "bin_height", check_int_range(
            bin_height, "bin_height", 1, 1_000_000_000_000))

    @staticmethod
    def from_packing_results(
            results: Iterable[PackingResult],
            collector: Callable[["PackingStatistics"], None]) -> None:
        """
        Create packing statistics from a sequence of packing results.

        :param results: the packing results
        :param collector: the collector receiving the created packing
            statistics
        """
        if not isinstance(results, Iterable):
            raise type_error(results, "results", Iterable)
        if not callable(collector):
            raise type_error(collector, "collector", call=True)
        groups: Final[dict[tuple[str, str, str], list[PackingResult]]] = {}
        objectives_set: set[str] = set()
        for i, pr in enumerate(results):
            if not isinstance(pr, PackingResult):
                raise type_error(pr, f"end_results[{i}]", PackingResult)
            setting: tuple[str, str, str] = \
                pr.end_result.algorithm, pr.end_result.instance, \
                pr.used_objective
            if setting in groups:
                groups[setting].append(pr)
            else:
                groups[setting] = [pr]
            objectives_set.update(pr.objectives.keys())

        if len(groups) <= 0:
            raise ValueError("results is empty!")
        if len(objectives_set) <= 0:
            raise ValueError("results has not objectives!")
        end_stats: Final[list[EndStatistics]] = []
        objectives: Final[list[str]] = sorted(objectives_set)

        for key in sorted(groups.keys()):
            data = groups[key]

            pr0 = data[0]
            n_items: int = pr0.n_items
            n_different_items: int = pr0.n_different_items
            bin_width: int = pr0.bin_width
            bin_height: int = pr0.bin_height
            used_objective: str = pr0.used_objective
            if used_objective not in objectives_set:
                raise ValueError(
                    f"{used_objective!r} not in {objectives_set!r}.")
            if used_objective != key[2]:
                raise ValueError(
                    f"used_objective={used_objective!r} different "
                    f"from key={key[2]}!?")
            objective_bounds: Mapping[str, int | float] = pr0.objective_bounds
            bin_bounds: Mapping[str, int] = pr0.bin_bounds
            for i, pr in enumerate(data):
                if n_items != pr.n_items:
                    raise ValueError(f"n_items={n_items} for data[0] but "
                                     f"{pr.n_items} for data[{i}]?")
                if n_different_items != pr.n_different_items:
                    raise ValueError(
                        f"n_different_items={n_different_items} for data[0] "
                        f"but {pr.n_different_items} for data[{i}]?")
                if bin_width != pr.bin_width:
                    raise ValueError(
                        f"bin_width={bin_width} for data[0] "
                        f"but {pr.bin_width} for data[{i}]?")
                if bin_height != pr.bin_height:
                    raise ValueError(
                        f"bin_height={bin_height} for data[0] "
                        f"but {pr.bin_height} for data[{i}]?")
                if used_objective != pr.used_objective:
                    raise ValueError(
                        f"used_objective={used_objective!r} for data[0] "
                        f"but {pr.used_objective!r} for data[{i}]?")
                if objective_bounds != pr.objective_bounds:
                    raise ValueError(
                        f"objective_bounds={objective_bounds!r} for data[0] "
                        f"but {pr.objective_bounds!r} for data[{i}]?")
                if bin_bounds != pr.bin_bounds:
                    raise ValueError(
                        f"bin_bounds={bin_bounds!r} for data[0] "
                        f"but {pr.bin_bounds!r} for data[{i}]?")

            EndStatistics.from_end_results((pr.end_result for pr in data),
                                           end_stats.append)
            if len(end_stats) != 1:
                raise ValueError(f"got {end_stats} from {data}?")

            collector(PackingStatistics(
                n_items=n_items,
                n_different_items=n_different_items,
                bin_width=bin_width,
                bin_height=bin_height,
                end_statistics=end_stats[0],
                objectives={
                    o: Statistics.create([pr.objectives[o] for pr in data])
                    for o in objectives
                },
                used_objective=used_objective,
                objective_bounds=objective_bounds,
                bin_bounds=bin_bounds,
            ))
            end_stats.clear()

    @staticmethod
    def to_csv(results: Iterable["PackingStatistics"], file: str) -> Path:
        """
        Write a sequence of packing statistics to a file in CSV format.

        :param results: the end statistics
        :param file: the path
        :return: the path of the file that was written
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing packing results to CSV file {path!r}.")
        Path.path(os.path.dirname(path)).ensure_dir_exists()

        # get a nicely sorted view on the statistics
        use_stats = sorted(
            results, key=lambda ppr: (
                ppr.end_statistics.algorithm,
                ppr.end_statistics.instance,
                ppr.used_objective,
                ppr.end_statistics))

        has_goal_f: int = 0  # 1
        has_best_f_scaled: bool = False  # 2
        has_n_success: bool = False  # 4
        has_success_fes: bool = False  # 8
        has_success_time_millis: bool = False  # 16
        has_ert_fes: bool = False  # 32
        has_ert_time_millis: bool = False  # 64
        has_max_fes: int = 0  # 128
        has_max_time_millis: int = 0  # 256
        checker: int = 511

        for ess in use_stats:
            es = ess.end_statistics
            if es.goal_f is not None:
                if isinstance(es.goal_f, Statistics):
                    has_goal_f = 2
                    checker &= ~1
                elif has_goal_f == 0:
                    has_goal_f = 1
            if es.best_f_scaled is not None:
                has_best_f_scaled = True
                checker &= ~2
            if es.n_success is not None:
                has_n_success = True
                checker &= ~4
            if es.success_fes is not None:
                has_success_fes = True
                checker &= ~8
            if es.success_time_millis is not None:
                has_success_time_millis = True
                checker &= ~16
            if es.ert_fes is not None:
                has_ert_fes = True
                checker &= ~32
            if es.ert_time_millis is not None:
                has_ert_time_millis = True
                checker &= ~64
            if es.max_fes is not None:
                if isinstance(es.max_fes, Statistics):
                    has_max_fes = 2
                    checker &= ~128
                elif has_max_fes == 0:
                    has_max_fes = 1
            if es.max_time_millis is not None:
                if isinstance(es.max_time_millis, Statistics):
                    has_max_time_millis = 2
                    checker &= ~256
                elif has_max_time_millis == 0:
                    has_max_time_millis = 1
            if checker == 0:
                break

        # get the names of the bounds and objectives
        bin_bounds_set: set[str] = set()
        objectives_set: set[str] = set()
        for pr in use_stats:
            bin_bounds_set.update(pr.bin_bounds.keys())
            objectives_set.update(pr.objectives.keys())
        bin_bounds: Final[list[str]] = sorted(bin_bounds_set)
        objectives: Final[list[str]] = sorted(objectives_set)

        with path.open_for_write() as out:
            wrt: Final[Callable] = out.write
            sep: Final[str] = CSV_SEPARATOR
            wrt(KEY_ALGORITHM)
            wrt(sep)
            wrt(KEY_INSTANCE)
            wrt(sep)
            wrt(KEY_USED_OBJECTIVE)
            wrt(sep)
            wrt(KEY_N_ITEMS)
            wrt(sep)
            wrt(KEY_N_DIFFERENT_ITEMS)
            wrt(sep)
            wrt(KEY_BIN_WIDTH)
            wrt(sep)
            wrt(KEY_BIN_HEIGHT)
            wrt(sep)
            for bb in bin_bounds:
                wrt(bb)
                wrt(sep)
            for oo in objectives:
                wrt(oo)
                wrt(_OBJECTIVE_LOWER)
                wrt(sep)
                wrt(oo)
                wrt(_OBJECTIVE_UPPER)
                wrt(sep)

            def h(p) -> None:
                wrt(sep.join(Statistics.csv_col_names(p)))

            wrt(KEY_N)
            wrt(sep)
            h(KEY_BEST_F)
            wrt(sep)
            h(KEY_LAST_IMPROVEMENT_FE)
            wrt(sep)
            h(KEY_LAST_IMPROVEMENT_TIME_MILLIS)
            wrt(sep)
            h(KEY_TOTAL_FES)
            wrt(sep)
            h(KEY_TOTAL_TIME_MILLIS)
            if has_goal_f == 1:
                wrt(sep)
                wrt(KEY_GOAL_F)
            elif has_goal_f == 2:
                wrt(sep)
                h(KEY_GOAL_F)
            if has_best_f_scaled:
                wrt(sep)
                h(KEY_BEST_F_SCALED)
            if has_n_success:
                wrt(sep)
                wrt(KEY_N_SUCCESS)
            if has_success_fes:
                wrt(sep)
                h(KEY_SUCCESS_FES)
            if has_success_time_millis:
                wrt(sep)
                h(KEY_SUCCESS_TIME_MILLIS)
            if has_ert_fes:
                wrt(sep)
                wrt(KEY_ERT_FES)
            if has_ert_time_millis:
                wrt(sep)
                wrt(KEY_ERT_TIME_MILLIS)
            if has_max_fes == 1:
                wrt(sep)
                wrt(KEY_MAX_FES)
            elif has_max_fes == 2:
                wrt(sep)
                h(KEY_MAX_FES)
            if has_max_time_millis == 1:
                wrt(sep)
                wrt(KEY_MAX_TIME_MILLIS)
            elif has_max_time_millis == 2:
                wrt(sep)
                h(KEY_MAX_TIME_MILLIS)
            for oo in objectives:
                wrt(sep)
                h(oo)

            out.write("\n")

            csv: Final[Callable] = Statistics.value_to_csv
            num: Final[Callable] = num_to_str

            for rec in use_stats:
                er = rec.end_statistics
                wrt(er.algorithm)
                wrt(sep)
                wrt(er.instance)
                wrt(sep)
                wrt(rec.used_objective)
                wrt(sep)
                wrt(str(rec.n_items))
                wrt(sep)
                wrt(str(rec.n_different_items))
                wrt(sep)
                wrt(str(rec.bin_width))
                wrt(sep)
                wrt(str(rec.bin_height))
                wrt(sep)

                for bb in bin_bounds:
                    wrt(str(rec.bin_bounds[bb]))
                    wrt(sep)
                for oo in objectives:
                    wrt(num_to_str(rec.objective_bounds[
                        f"{oo}{_OBJECTIVE_LOWER}"]))
                    wrt(sep)
                    wrt(num_to_str(rec.objective_bounds[
                        f"{oo}{_OBJECTIVE_UPPER}"]))
                    wrt(sep)

                wrt(str(er.n))
                wrt(sep)
                wrt(er.best_f.to_csv())
                wrt(sep)
                wrt(er.last_improvement_fe.to_csv())
                wrt(sep)
                wrt(er.last_improvement_time_millis.to_csv())
                wrt(sep)
                wrt(er.total_fes.to_csv())
                wrt(sep)
                wrt(er.total_time_millis.to_csv())
                if has_goal_f == 1:
                    wrt(sep)
                    if er.goal_f is not None:
                        wrt(num(er.goal_f))
                elif has_goal_f == 2:
                    wrt(sep)
                    if isinstance(er.goal_f, Statistics):
                        wrt(er.goal_f.to_csv())
                    elif isinstance(er.goal_f, int | float):
                        wrt(csv(er.goal_f))
                    else:
                        wrt(EMPTY_CSV_ROW)
                if has_best_f_scaled:
                    wrt(sep)
                    if er.best_f_scaled is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.best_f_scaled.to_csv())
                if has_n_success:
                    wrt(sep)
                    if er.n_success is not None:
                        wrt(str(er.n_success))
                if has_success_fes:
                    wrt(sep)
                    if er.success_fes is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.success_fes.to_csv())
                if has_success_time_millis:
                    wrt(sep)
                    if er.success_time_millis is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.success_time_millis.to_csv())
                if has_ert_fes:
                    wrt(sep)
                    if er.ert_fes is not None:
                        wrt(num(er.ert_fes))
                if has_ert_time_millis:
                    wrt(sep)
                    if er.ert_time_millis is not None:
                        wrt(num(er.ert_time_millis))
                if has_max_fes == 1:
                    wrt(sep)
                    if er.max_fes is not None:
                        wrt(str(er.max_fes))
                elif has_max_fes == 2:
                    wrt(sep)
                    if isinstance(er.max_fes, Statistics):
                        wrt(er.max_fes.to_csv())
                    elif isinstance(er.max_fes, int | float):
                        wrt(csv(er.max_fes))
                    else:
                        wrt(EMPTY_CSV_ROW)
                if has_max_time_millis == 1:
                    wrt(sep)
                    if er.max_time_millis is not None:
                        wrt(str(er.max_time_millis))
                elif has_max_time_millis == 2:
                    wrt(sep)
                    if isinstance(er.max_time_millis, Statistics):
                        wrt(er.max_time_millis.to_csv())
                    elif isinstance(er.max_time_millis, int | float):
                        wrt(csv(er.max_time_millis))
                    else:
                        wrt(EMPTY_CSV_ROW)
                for oo in objectives:
                    wrt(sep)
                    wrt(rec.objectives[oo].to_csv())
                out.write("\n")

        # finally, return the path to the generated file
        return path



"""Synthesize some interesting starting points."""

from typing import Callable, Final, Iterable, cast

import numba  # type: ignore
import numpy as np
from moptipy.algorithms.so.vector.cmaes_lib import BiPopCMAES
from moptipy.api.execution import Execution
from moptipy.api.objective import Objective
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.console import logger


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def interesting_point_transform(
        x: np.ndarray, max_radius: float, dim: int) -> np.ndarray:
    """
    Transform interesting points.

    >>> xx = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12.0])
    >>> ppp = interesting_point_transform(xx, 10.0, 3)
    >>> print(ppp)
    [[0.6681531  1.33630621 2.00445931]
     [2.27921153 2.84901441 3.41881729]
     [3.76928033 4.30774895 4.84621757]
     [5.23423923 5.75766315 6.28108707]]
    >>> print(xx)
    [0.6681531  1.33630621 2.00445931 2.27921153 2.84901441 3.41881729
     3.76928033 4.30774895 4.84621757 5.23423923 5.75766315 6.28108707]
    >>> print([np.sqrt(np.square(pppp).sum()) for pppp in ppp])
    [2.5, 5.0, 7.5, 10.0]
    >>> ppp = interesting_point_transform(xx, 10.0, 3)
    >>> print(ppp)
    [[0.6681531  1.33630621 2.00445931]
     [2.27921153 2.84901441 3.41881729]
     [3.76928033 4.30774895 4.84621757]
     [5.23423923 5.75766315 6.28108707]]
    >>> print([np.sqrt(np.square(pppp).sum()) for pppp in ppp])
    [2.5, 5.0, 7.5, 10.0]
    >>> ppp = interesting_point_transform(xx, 10.0, 2)
    >>> print(ppp)
    [[0.74535599 1.49071198]
     [2.20132119 2.50305736]
     [3.200922   3.8411064 ]
     [4.39003072 5.01717796]
     [5.66154475 6.11484713]
     [6.75724629 7.3715414 ]]
    """
    n: Final[int] = len(x) // dim
    p: Final[np.ndarray] = np.reshape(x, (n, dim))
    for i in range(n):
        pp: np.ndarray = p[i, :]

        cur_radius: float = np.sqrt(np.square(pp).sum())
        if cur_radius <= 0.0:
            continue
        goal_radius: float = max_radius * ((i + 1) / n)
        if goal_radius != cur_radius:
            first: bool = True
            while True:
                old_radius = cur_radius
                pp2 = pp * goal_radius / cur_radius
                if np.array_equal(pp2, pp):
                    break
                cur_radius = np.sqrt(np.square(pp2).sum())
                if cur_radius == goal_radius:
                    pp = pp2
                    break
                if first:
                    first = False
                elif cur_radius > old_radius:
                    break
                pp = pp2
        p[i, :] = pp
    return p


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def interesting_point_objective(x: np.ndarray, other: np.ndarray,
                                max_radius: float, dim: int) -> float:
    """Compute the point diversity."""
    pts: np.ndarray = interesting_point_transform(x, max_radius, dim)
    n: Final[int] = len(pts)
    f: float = 0.0
    for i in range(n):
        pt: np.ndarray = pts[i]
        scale: float = (((i + 1) / n) ** 2)  # we know this
        for j in range(i + 1, n):
            dst: float = np.sqrt(np.square(pt - pts[j]).sum())
            f += scale / dst if dst > 0.0 else 1e10
        for oth in other:
            dst = np.sqrt(np.square(pt - oth).sum())
            f += scale / dst if dst > 0.0 else 1e10
    return f


def make_interesting_starting_points(
        n: int, other: Iterable[Iterable[float]],
        log: bool = True) -> np.ndarray:
    """
    Create some reasonably diverse starting points.

    :param n: the number of starting points
    :param other: the other points
    :param log: write log output
    :return: the starting points

    >>> p = make_interesting_starting_points(
    ...     3, np.array([[1.0, 2.0], [3.0, 2.9]]), False)
    >>> print(",".join(";".join(f"{x:.5f}" for x in row) for row in p))
    0.92722;-1.55225,-3.61077;0.19795,-0.91540;-5.34649

    >>> p = make_interesting_starting_points(
    ...     3, np.array([[1.0, 2.0, 7.0], [3.0, 2.9, 1.1]]), False)
    >>> print(",".join(";".join(f"{x:.5f}" for x in row) for row in p))
    -0.35688;-3.08116;0.72047,-5.49372;1.93835;-2.57330,0.22616;-7.21443;-6.25786
    """
    other_points: Final[np.ndarray] = np.array(other)
    dim: Final[int] = other_points.shape[1]
    max_fes: Final[int] = 2048 + int(40 * (n ** (dim / 1.7)))

    max_radius: float = max(np.sqrt(np.square(pt).sum())
                            for pt in other_points)
    max_radius = max(max_radius ** 1.1, max_radius ** 0.6,
                     max_radius * 1.3)

    max_dim: float = np.max(np.abs(
        np.array(other_points).flatten()))
    max_dim = max(max_dim ** 1.2, max_dim ** (1.0 / 1.2),
                  max_dim ** 2.5, max_radius * 1.2)

    if log:
        logger(
            f"now determining {n} hopefully diverse samples of dimension "
            f"{dim} using {len(other_points)} other points, "
            f"max_radius={max_radius}, and max_dim={max_dim} for "
            f"{max_fes} FEs.")

    space: Final[VectorSpace] = VectorSpace(dim * n, -max_dim, max_dim)
    best: np.ndarray = space.create()

    class __Obj(Objective):
        def __init__(self):
            nonlocal other_points
            nonlocal max_radius
            nonlocal dim
            self.evaluate = cast(  # type: ignore
                Callable[[np.ndarray], float],
                lambda x, o=other_points, mr=max_radius, dd=dim:
                interesting_point_objective(x, o, mr, dd))

    with Execution().set_solution_space(space).set_rand_seed(1)\
            .set_algorithm(BiPopCMAES(space))\
            .set_max_fes(max_fes)\
            .set_objective(__Obj()).execute() as process:
        f = process.get_best_f()
        process.get_copy_of_best_x(best)
        best = interesting_point_transform(best, max_radius, dim)
        if log:
            logger(f"generated {n} points with objective {f}:\n{best!r}")
    return best



"""A base class for implementing controllers."""


from typing import Callable, Final

import numpy as np
from moptipy.api.component import Component
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import check_to_int_range, type_error


class Controller(Component):
    """A class for governing a system via differential system."""

    def __init__(self, name: str,
                 state_dims: int, control_dims: int, param_dims: int,
                 func: Callable[[np.ndarray, float, np.ndarray,
                                 np.ndarray], None] | None = None) -> None:
        """
        Initialize the system.

        :param name: the name of the system.
        :param state_dims: the state dimensions
        :param control_dims: the control dimensions
        :param param_dims: the parameter dimensions
        """
        super().__init__()
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        #: the controller name
        self.name: Final[str] = name
        #: the dimensions of the state variable
        self.state_dims: Final[int] = check_to_int_range(
            state_dims, "state_dims", 2, 3)
        #: the dimensions of the controller output
        self.control_dims: Final[int] = check_to_int_range(
            control_dims, "control_dims", 1, 100)
        #: the dimensions of the controller parameter
        self.param_dims: Final[int] = check_to_int_range(
            param_dims, "param_dims", 1, 1_000)
        if func is not None:
            if not callable(func):
                raise type_error(func, "func", None, call=True)
            self.controller = func  # type: ignore

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("stateDims", self.state_dims)
        logger.key_value("controlDims", self.control_dims)
        logger.key_value("paramDims", self.param_dims)

    def __str__(self):
        """
        Get the name of this controller.

        :return: the name of this controller
        """
        return self.name

    def controller(self, state: np.ndarray,  # pylint: disable=E0202
                   time: float,  # pylint: disable=E0202
                   params: np.ndarray,  # pylint: disable=E0202
                   out: np.ndarray) -> None:  # pylint: disable=E0202
        """
        Compute the control value and store it in `out`.

        :param state: the state vector
        :param time: the time value
        :param params: the controller variables
        :param out: the output array to receive the controller values
        """

    def parameter_space(self) -> VectorSpace:
        """
        Create a vector space to represent the possible parameterizations.

        :return: a vector space for the possible parameterizations of this
            controller.
        """
        return VectorSpace(self.param_dims, -32.0, 32.0)



"""An instance of the dynamic control problem."""

from typing import Final

import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.path import Path
from moptipy.utils.types import type_error

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


class Instance(Component):
    """An instance of the dynamic control problem."""

    def __init__(self, system: System, controller: Controller) -> None:
        """
        Create an instance of the dynamic control problem.

        :param system: the system of equations governing the dynamic system
        :param controller: the controller applied to the system
        """
        super().__init__()
        if not isinstance(system, System):
            raise type_error(system, "system", System)
        if not isinstance(controller, Controller):
            raise type_error(controller, "controller", Controller)
        if controller.state_dims != system.state_dims:
            raise ValueError(
                f"controller.state_dims={controller.state_dims}, but "
                f"system.state_dims={system.state_dims} for controller "
                f"{str(controller)!r} and system {str(system)!r}.")
        if controller.control_dims != system.control_dims:
            raise ValueError(
                f"controller.control_dims={controller.control_dims}, but "
                f"system.control_dims={system.control_dims} for controller "
                f"{str(controller)!r} and system {str(system)!r}.")
        #: the system governing the dynamic system
        self.system: Final[System] = system
        #: the controller applied to the system
        self.controller: Final[Controller] = controller

    def __str__(self) -> str:
        """
        Get the name of this instance.

        :return: a combination of the equation name and the controller name
        """
        return f"{self.system}_{self.controller}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("sys") as scope:
            self.system.log_parameters_to(scope)
        with logger.scope("ctrl") as scope:
            self.controller.log_parameters_to(scope)

    def describe_parameterization(
            self, title: str | None,
            parameters: np.ndarray, base_name: str,
            dest_dir: str) -> tuple[Path, Path]:
        """
        Describe the performance of a given system of system.

        :param title: the optional title
        :param parameters: the controller parameters
        :param base_name: the base name of the file to produce
        :param dest_dir: the destination directory
        :returns: the paths of the generated files
        """
        the_title = self.controller.name
        if title is not None:
            the_title = f"{the_title}\n{title}"
        return self.system.describe_system(
            the_title, self.controller.controller, parameters,
            base_name, dest_dir)
